# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import base_compute
import base_network
import base_storage

from keystoneclient import exceptions as keystone_exception
import log as logging
import sys
import users

LOG = logging.getLogger(__name__)

class KBFlavorCheckException(Exception):
    pass

class KBQuotaCheckException(Exception):
    pass

class Tenant(object):
    """
    Holds the tenant resources
    1. Provides ability to create users in a tenant
    2. Uses the User class to perform all user resource creation and deletion
    """

    def __init__(self, tenant_name, kloud, tenant_quota, reusing_users=None):
        """
        Holds the tenant name
        tenant id and keystone client
        Also stores the auth_url for constructing credentials
        Stores the shared network in case of testing and
        tested cloud being on same cloud
        """
        self.kloud = kloud
        self.res_logger = kloud.res_logger
        self.tenant_name = tenant_name
        self.tenant_api = self.kloud.keystone.tenants \
            if self.kloud.keystone.version == 'v2.0' else self.kloud.keystone.projects
        if not self.kloud.reusing_tenants:
            self.tenant_object = self._get_tenant()
            self.tenant_id = self.tenant_object.id
        else:
            LOG.info("Using tenant: " + self.tenant_name)
            # Only admin can retrive the object via Keystone API
            self.tenant_object = None
            self.tenant_id = self.kloud.keystone.tenant_id

        self.tenant_quota = tenant_quota
        self.reusing_users = reusing_users
        # Contains a list of user instance objects
        self.user_list = []

    def _get_tenant(self):

        '''
        Create or reuse a tenant object of a given name
        '''
        try:
            LOG.info("Creating tenant: " + self.tenant_name)
            tenant_object = \
                self.tenant_api.create(self.tenant_name,
                                       # domain="default",
                                       description="KloudBuster tenant",
                                       enabled=True)
            return tenant_object
        except keystone_exception.Conflict as exc:
            # ost likely the entry already exists:
            # Conflict: Conflict occurred attempting to store project - Duplicate Entry (HTTP 409)
            if exc.http_status != 409:
                raise exc
            LOG.info("Tenant %s already present, reusing it" % self.tenant_name)
            return self.tenant_api.find(name=self.tenant_name)

        # Should never come here
        raise Exception()

    def update_quota(self):
        nova_quota = base_compute.NovaQuota(self.kloud.nova_client, self.tenant_id)
        nova_quota.update_quota(**self.tenant_quota['nova'])

        if self.kloud.storage_mode:
            cinder_quota = base_storage.CinderQuota(self.kloud.cinder_client, self.tenant_id)
            cinder_quota.update_quota(**self.tenant_quota['cinder'])

        neutron_quota = base_network.NeutronQuota(self.kloud.neutron_client, self.tenant_id)
        neutron_quota.update_quota(self.tenant_quota['neutron'])

    def check_quota(self):
        # Flavor check
        flavor_manager = base_compute.Flavor(self.kloud.nova_client)
        find_flag = False
        fcand = {'vcpus': sys.maxint, 'ram': sys.maxint, 'disk': sys.maxint}
        for flavor in flavor_manager.list():
            flavor = vars(flavor)
            if flavor['vcpus'] < 1 or flavor['ram'] < 1024 or flavor['disk'] < 10:
                continue
            if flavor['vcpus'] < fcand['vcpus']:
                fcand = flavor
            if flavor['vcpus'] == fcand['vcpus'] and flavor['ram'] < fcand['ram']:
                fcand = flavor
            if flavor['vcpus'] == fcand['vcpus'] and flavor['ram'] == fcand['ram'] and\
               flavor['disk'] < fcand['disk']:
                fcand = flavor
            find_flag = True

        if find_flag:
            LOG.info('Automatically selects flavor %s to instantiate VMs.' % fcand['name'])
            self.kloud.flavor_to_use = fcand['name']
        else:
            LOG.error('Cannot find a flavor which meets the minimum '
                      'requirements to instantiate VMs.')
            raise KBFlavorCheckException()

        # Nova/Cinder/Neutron quota check
        tenant_id = self.tenant_id
        meet_quota = True
        for quota_type in ['nova', 'cinder', 'neutron']:
            if quota_type == 'nova':
                quota_manager = base_compute.NovaQuota(self.kloud.nova_client, tenant_id)
            elif quota_type == 'cinder':
                quota_manager = base_storage.CinderQuota(self.kloud.cinder_client, tenant_id)
            else:
                quota_manager = base_network.NeutronQuota(self.kloud.neutron_client, tenant_id)

            meet_quota = True
            quota = quota_manager.get()
            for key, value in self.tenant_quota[quota_type].iteritems():
                if quota[key] < value:
                    meet_quota = False
                    break

        if not meet_quota:
            LOG.error('%s quota is too small. Minimum requirement: %s.' %
                      (quota_type, self.tenant_quota[quota_type]))
            raise KBQuotaCheckException()

    def create_resources(self):
        """
        Creates all the entities associated with
        a user offloads tasks to user class
        """
        if self.kloud.reusing_tenants:
            self.check_quota()
        else:
            self.update_quota()

        if self.reusing_users:
            user_name = self.reusing_users['username']
            password = self.reusing_users['password']
            user_instance = users.User(user_name, password, self, '_member_')
            self.user_list.append(user_instance)
        else:
            user_name = self.tenant_name + "-U"
            user_instance = users.User(user_name, user_name, self,
                                       self.kloud.scale_cfg['keystone_admin_role'])
            # Global list with all user instances
            self.user_list.append(user_instance)
            self.res_logger.log('users', user_instance.user_name, user_instance.user.id)

        for user_instance in self.user_list:
            # Now create the user resources like routers which inturn trigger network and
            # vm creation
            user_instance.create_resources()

    def get_first_network(self):
        if self.user_list:
            return self.user_list[0].get_first_network()
        return None

    def get_all_instances(self):
        all_instances = []
        for user in self.user_list:
            all_instances.extend(user.get_all_instances())
        return all_instances

    def get_prefix(self):
        return self.kloud.get_prefix() + '_' + self.prefix

    def delete_resources(self):
        """
        Delete all user resources and than
        deletes the tenant
        """
        flag = True
        # Delete all the users in the tenant along with network and compute elements
        for user in self.user_list:
            flag = flag & user.delete_resources()

        if not self.reusing_users:
            # Delete the tenant (self)
            self.tenant_api.delete(self.tenant_id)

        return flag



# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import json

from perf_tool import PerfTool

from hdrh.histogram import HdrHistogram
import log as logging

LOG = logging.getLogger(__name__)


class WrkTool(PerfTool):

    def __init__(self, instance):
        PerfTool.__init__(self, instance, 'wrk2')

    def cmd_parser_run_client(self, status, stdout, stderr):
        if status:
            return self.parse_error(stderr)

        # Sample Output:
        # {
        # "total_req": 29, "rps": 29.39, "rx_bps": "1.43MB",
        # "hist": "HISTggAAANB4nO3XsQ3DIBAAQFBWcJnskc6reJMokjfIcF7DI9iS3bkIFBg"
        #         "k7iRE8wU88Hqe8+8b4ucdDo9zjvsYXssUxjUAAAAAAAAAAAAAAAAAAAAAAA"
        #         "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
        #         "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAQKpYKS5V6+sDAACu7u7P/Rvq"
        #         "kKe65L9tuefTaj2EnngXdeXWr1L17l98r/ek1323plR/ETdacAUk"
        # "errors": {"read": 1},
        # }

        try:
            result = json.loads(stdout)
            http_total_req = int(result['total_req'])
            http_rps = float(result['rps'])
            http_tp_kbytes = result['rx_bps']
            # Uniform in unit MB
            ex_unit = 'KMG'.find(http_tp_kbytes[-2])
            if ex_unit == -1:
                raise ValueError
            val = float(http_tp_kbytes[0:-2])
            http_tp_kbytes = float(val * (1024 ** (ex_unit)))

            if 'errors' in result:
                errs = []
                for key in ['connect', 'read', 'write', 'timeout', 'http_error']:
                    if key in result['errors']:
                        errs.append(int(result['errors'][key]))
                    else:
                        errs.append(0)
                http_sock_err = errs[0] + errs[1] + errs[2]
                http_sock_timeout = errs[3]
                http_err = errs[4]
            else:
                http_sock_err = 0
                http_sock_timeout = 0
                http_err = 0

            latency_stats = result['hist']
        except Exception:
            return self.parse_error('Could not parse: "%s"' % (stdout))

        parsed_output = {'tool': self.name}
        if http_total_req:
            parsed_output['http_total_req'] = http_total_req
        if http_rps:
            parsed_output['http_rps'] = http_rps
        if http_tp_kbytes:
            parsed_output['http_throughput_kbytes'] = http_tp_kbytes
        if http_sock_err:
            parsed_output['http_sock_err'] = http_sock_err
        if http_sock_timeout:
            parsed_output['http_sock_timeout'] = http_sock_timeout
        if http_err:
            parsed_output['http_err'] = http_err
        if latency_stats:
            parsed_output['latency_stats'] = latency_stats

        return parsed_output

    @staticmethod
    def consolidate_results(results):
        err_flag = False
        all_res = {'tool': 'wrk2'}
        total_count = len(results)
        if not total_count:
            return all_res

        for key in ['http_rps', 'http_total_req', 'http_sock_err',
                    'http_sock_timeout', 'http_throughput_kbytes']:
            all_res[key] = 0
            for item in results:
                all_res[key] += item['results'].get(key, 0)
            all_res[key] = int(all_res[key])

        if 'latency_stats' in results[0]['results']:
            # for item in results:
            #     print item['results']['latency_stats']
            all_res['latency_stats'] = []
            histogram = HdrHistogram(1, 24 * 3600 * 1000 * 1000, 2)
            for item in results:
                if 'latency_stats' in item['results']:
                    histogram.decode_and_add(item['results']['latency_stats'])
                else:
                    err_flag = True
            perc_list = [50, 75, 90, 99, 99.9, 99.99, 99.999]
            latency_dict = histogram.get_percentile_to_value_dict(perc_list)
            for key, value in latency_dict.iteritems():
                all_res['latency_stats'].append([key, value])
            all_res['latency_stats'].sort()

        if err_flag:
            LOG.warning('Unable to find latency_stats from the result dictionary, this '
                        'may indicate that the test application on VM exited abnormally.')

        return all_res

    @staticmethod
    def consolidate_samples(results, vm_count):
        all_res = WrkTool.consolidate_results(results)
        total_count = float(len(results)) / vm_count
        if not total_count:
            return all_res

        all_res['http_rps'] = int(all_res['http_rps'] / total_count)
        all_res['http_throughput_kbytes'] = int(all_res['http_throughput_kbytes'] / total_count)
        return all_res



# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import division

from kb_runner_base import KBException
from kb_runner_base import KBRunner
import log as logging

LOG = logging.getLogger(__name__)

class KBSetStaticRouteException(KBException):
    pass

class KBHTTPServerUpException(KBException):
    pass

class KBHTTPBenchException(KBException):
    pass

class KBRunner_HTTP(KBRunner):
    """
    Control the testing VMs on the testing cloud
    """

    def __init__(self, client_list, config, expected_agent_version, single_cloud=True):
        KBRunner.__init__(self, client_list, config, expected_agent_version, single_cloud)

    def header_formatter(self, stage, vm_count):
        conns = vm_count * self.config.http_tool_configs.connections
        rate_limit = vm_count * self.config.http_tool_configs.rate_limit
        msg = "Stage %d: %d VM(s), %d Connections, %d Expected RPS" %\
              (stage, vm_count, conns, rate_limit)
        return msg

    def setup_static_route(self, active_range, timeout=30):
        func = {'cmd': 'setup_static_route', 'active_range': active_range}
        self.send_cmd('EXEC', 'http', func)
        cnt_succ = self.polling_vms(timeout)[0]
        if cnt_succ != len(self.client_dict):
            raise KBSetStaticRouteException()

    def check_http_service(self, active_range, timeout=30):
        func = {'cmd': 'check_http_service', 'active_range': active_range}
        self.send_cmd('EXEC', 'http', func)
        cnt_succ = self.polling_vms(timeout)[0]
        if cnt_succ != len(self.client_dict):
            raise KBHTTPServerUpException()

    def run_http_test(self, active_range):
        func = {'cmd': 'run_http_test', 'active_range': active_range,
                'parameter': dict(self.config.http_tool_configs)}
        self.send_cmd('EXEC', 'http', func)
        # Give additional 30 seconds for everybody to report results
        timeout = self.config.http_tool_configs.duration + 30
        cnt_pending = self.polling_vms(timeout)[2]
        if cnt_pending != 0:
            LOG.warning("Testing VMs are not returning results within grace period, "
                        "summary shown below may not be accurate!")

        # Parse the results from HTTP benchmarking tool
        for key, instance in self.client_dict.items():
            self.result[key] = instance.perf_client_parser(**self.result[key])

    def single_run(self, active_range=None, http_test_only=False):
        try:
            if not http_test_only:
                if self.single_cloud:
                    LOG.info("Setting up static route to reach tested cloud...")
                    self.setup_static_route(active_range)

                LOG.info("Waiting for HTTP service to come up...")
                self.check_http_service(active_range)

                if self.config.prompt_before_run:
                    print "Press enter to start running benchmarking tools..."
                    raw_input()

            LOG.info("Running HTTP Benchmarking...")
            self.report = {'seq': 0, 'report': None}
            self.result = {}
            self.run_http_test(active_range)

            # Call the method in corresponding tools to consolidate results
            perf_tool = self.client_dict.values()[0].perf_tool
            LOG.kbdebug(self.result.values())
            self.tool_result = perf_tool.consolidate_results(self.result.values())
            self.tool_result['http_rate_limit'] =\
                len(self.client_dict) * self.config.http_tool_configs.rate_limit
            self.tool_result['total_connections'] =\
                len(self.client_dict) * self.config.http_tool_configs.connections
            vm_count = active_range[1] - active_range[0] + 1\
                if active_range else len(self.full_client_dict)
            self.tool_result['total_client_vms'] = vm_count
            self.tool_result['total_server_vms'] = self.tool_result['total_client_vms']
            # self.tool_result['host_stats'] = self.gen_host_stats()
        except KBSetStaticRouteException:
            raise KBException("Could not set static route.")
        except KBHTTPServerUpException:
            raise KBException("HTTP service is not up in testing cloud.")
        except KBHTTPBenchException:
            raise KBException("Error while running HTTP benchmarking tool.")

    def run(self, test_only=False):
        if not test_only:
            # Resources are already staged, just re-run the HTTP benchmarking tool
            self.wait_for_vm_up()

        if self.config.progression.enabled:
            self.tool_result = {}
            start = self.config.progression.vm_start
            multiple = self.config.progression.vm_multiple
            limit = self.config.progression.http_stop_limit
            timeout = self.config.http_tool_configs.timeout
            vm_list = self.full_client_dict.keys()
            vm_list.sort(cmp=lambda x, y: cmp(int(x[x.rfind('I') + 1:]), int(y[y.rfind('I') + 1:])))
            self.client_dict = {}
            cur_stage = 1

            while True:
                cur_vm_count = len(self.client_dict)
                if start == 1:
                    target_vm_count = 1 if cur_stage == 1 else (cur_stage - 1) * multiple
                else:
                    target_vm_count = cur_stage * multiple
                timeout_at_percentile = 0
                if target_vm_count > len(self.full_client_dict):
                    break
                if self.tool_result and 'latency_stats' in self.tool_result:
                    err = self.tool_result['http_sock_err'] + self.tool_result['http_sock_timeout']
                    pert_dict = dict(self.tool_result['latency_stats'])
                    if limit[1] in pert_dict.keys():
                        timeout_at_percentile = pert_dict[limit[1]] // 1000000
                    elif limit[1] != 0:
                        LOG.warning('Percentile %s%% is not a standard statistic point.' % limit[1])
                    if err > limit[0] or timeout_at_percentile > timeout:
                        LOG.warning('KloudBuster is stopping the iteration because the result '
                                    'reaches the stop limit.')
                        break

                for idx in xrange(cur_vm_count, target_vm_count):
                    self.client_dict[vm_list[idx]] = self.full_client_dict[vm_list[idx]]
                description = "-- %s --" % self.header_formatter(cur_stage, len(self.client_dict))
                LOG.info(description)
                self.single_run(active_range=[0, target_vm_count - 1],
                                http_test_only=test_only)
                LOG.info('-- Stage %s: %s --' % (cur_stage, str(self.tool_result)))
                self.tool_result['description'] = description
                cur_stage += 1
                yield self.tool_result
        else:
            self.single_run(http_test_only=test_only)
            yield self.tool_result



# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from hdrh.histogram import HdrHistogram
import json
import multiprocessing
import redis
import socket
import struct
import subprocess
from subprocess import Popen
import sys
import threading
import time
import traceback

# Define the version of the KloudBuster agent and VM image
#
# When VM is up running, the agent will send the READY message to the
# KloudBuster main program, along with its version. The main program
# will check the version to see whether the image meets the minimum
# requirements to run, and stopped with an error if not.
#
# This version must be incremented if the interface changes or if new features
# are added to the agent VM
__version__ = '6'

# TODO(Logging on Agent)

def get_image_name():
    '''Return the versioned VM image name that corresponds to this
    agent code. This string must match the way DIB names the kloudbuster image.
    Return:
        the versioned image name without the extension ('.qcow2' is implicit)
    '''
    return 'kloudbuster_v' + __version__

def get_image_version():
    return __version__

def exec_command(cmd, cwd=None):
    p = subprocess.Popen(cmd, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (stdout, stderr) = p.communicate()

    return p.returncode

def refresh_clock(clocks, force_sync=False):
    step = " "
    if force_sync:
        step = " -b "
    command = "sudo ntpdate" + step + clocks
    exec_command(command.split(" "))

class KB_Instance(object):

    # Check whether the HTTP Service is up running
    @staticmethod
    def check_http_service(target_url):
        cmd = 'while true; do\n'
        cmd += 'curl --head %s --connect-timeout 2 --silent\n' % (target_url)
        cmd += 'if [ $? -eq 0 ]; then break; fi\n'
        cmd += 'done'
        return cmd

    # Add static route
    @staticmethod
    def add_static_route(network, next_hop_ip, if_name=None):
        debug_msg = "Adding static route %s with next hop %s" % (network, next_hop_ip)
        cmd = "sudo ip route add %s via %s" % (network, next_hop_ip)
        if if_name:
            debug_msg += " and %s" % if_name
            cmd += " dev %s" % if_name
        print debug_msg
        return cmd

    @staticmethod
    def add_multicast_route():
        cmd = "sudo route add -net 224.0.0.0/8 dev eth0"
        return exec_command(cmd.split(" "))

    # Get static route
    @staticmethod
    def get_static_route(network, next_hop_ip=None, if_name=None):
        cmd = "ip route show %s" % network
        if next_hop_ip:
            cmd += " via %s" % next_hop_ip
        if if_name:
            cmd += " dev %s" % if_name
        return cmd

    # Delete static route
    @staticmethod
    def delete_static_route(network, next_hop_ip=None, if_name=None):
        debug_msg = "Deleting static route %s" % network
        cmd = "sudo ip route del %s" % network
        if next_hop_ip:
            debug_msg = " with next hop %s" % next_hop_ip
            cmd += " via %s" % next_hop_ip
        if if_name:
            if next_hop_ip:
                debug_msg = " and %s" % if_name
            else:
                debug_msg = "with next hop %s" % if_name
            cmd += " dev %s" % if_name
        print debug_msg
        return cmd

    # Run the HTTP benchmarking tool
    @staticmethod
    def run_wrk2(dest_path, target_url, threads, connections,
                 rate_limit, duration, timeout, connection_type,
                 report_interval):
        if not rate_limit:
            rate_limit = 65535

        cmd = '%s -t%d -c%d -R%d -d%ds -p%ds --timeout %ds -D2 -e %s' % \
              (dest_path, threads, connections, rate_limit, duration,
               report_interval, timeout, target_url)
        return cmd

    # Run the multicast benchmarking tool
    # -p$1:$2 => $1 is the local port, $2 is the multicast port
    # SOURCE PORT should be the same for any one address that you want to
    @staticmethod
    def run_nuttcp(dest_path, target_url, target_port, multicast_addr, multicast_port, bitrate=100,
                   pktsize=1316, report_interval=4, transmit_receive='t'):

        if bitrate < 1:
            bitrate = str(bitrate * 1000) + "k"
        else:
            bitrate = str(bitrate) + "m"

        cmd = '%s -%s -fxmitstats -j -o -fparse -m30 -Ri%s/%s -l%i -T%i -i1 -g%s -p%d -P%d %s' % \
            (dest_path, transmit_receive, bitrate, bitrate, pktsize, report_interval,
             multicast_addr, multicast_port, target_port, target_url)
        return cmd

    # Init volume
    @staticmethod
    def init_volume(dest_path, size, mkfs):
        cmd = 'if [ ! -e /kb_mnt ]; then\n'
        cmd += 'mkfs.xfs /dev/vdb && ' if mkfs else ''
        cmd += 'mkdir -p /kb_mnt && '
        cmd += 'mount /dev/vdb /kb_mnt && '
        cmd += '%s --name=create_file --filename=/kb_mnt/kb_storage_test.bin '\
               '--size=%s --create_only=1\n' % (dest_path, size)
        cmd += 'fi'
        return cmd

    # Run fio
    @staticmethod
    def run_fio(dest_path, name, description, mode, block_size, iodepth, runtime,
                rate_iops=None, rate=None, rwmixread=None, status_interval=None, extra_opts=None):
        fixed_opt = '--thread --ioengine=libaio --output-format=json+ --direct=1 '
        fixed_opt += '--filename=/kb_mnt/kb_storage_test.bin '
        required_opt = '--name=%s --rw=%s --bs=%s --iodepth=%s --runtime=%s ' %\
            (name, mode, block_size, iodepth, runtime)
        optional_opt = ''
        optional_opt += '--rate_iops=%s ' % rate_iops if rate_iops else ''
        optional_opt += '--rate=%s ' % rate if rate else ''
        optional_opt += '--rwmixread=%s ' % rwmixread if rwmixread else ''
        optional_opt += '--status-interval=%s ' % status_interval if status_interval else ''
        optional_opt += extra_opts if extra_opts else ''
        cmd = '%s %s %s %s' % (dest_path, fixed_opt, required_opt, optional_opt)
        return cmd

class KBA_Client(object):

    def __init__(self, user_data):
        host = user_data['redis_server']
        port = user_data['redis_server_port']
        self.user_data = user_data
        self.redis_obj = redis.StrictRedis(host=host, port=port)
        self.pubsub = self.redis_obj.pubsub(ignore_subscribe_messages=True)
        self.hello_thread = None
        self.stop_hello = threading.Event()
        self.vm_name = user_data['vm_name']
        self.orches_chan_name = "kloudbuster_orches"
        self.report_chan_name = "kloudbuster_report"
        self.last_cmd = None
        self.last_process = None

    def setup_channels(self):
        # Check for connections to redis server
        while (True):
            try:
                self.redis_obj.get("test")
            except (redis.exceptions.ConnectionError):
                time.sleep(1)
                continue
            break

        # Subscribe to orchestration channel
        self.pubsub.subscribe(self.orches_chan_name)

    def report(self, cmd, client_type, data):
        message = {'cmd': cmd, 'sender-id': self.vm_name,
                   'client-type': client_type, 'data': data}
        self.redis_obj.publish(self.report_chan_name, message)

    def send_hello(self):
        # Sending "hello" message to master node every 2 seconds
        while not self.stop_hello.is_set():
            self.report('READY', None, __version__)
            time.sleep(2)

    def post_processing(self, p_out):
        return p_out

    def exec_command(self, cmd):
        # Execute the command, and returns the outputs
        cmds = ['bash', '-c']
        cmds.append(cmd)
        p = subprocess.Popen(cmds, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        self.last_process = p
        (stdout, stderr) = p.communicate()

        return (p.returncode, stdout, stderr)

    def exec_command_report(self, cmd):
        # Execute the command, reporting periodically, and returns the outputs
        cmd_res_dict = None
        cmds = ['bash', '-c']
        cmds.append(cmd)
        p_output = ''
        p = subprocess.Popen(cmds, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        self.last_process = p
        lines_iterator = iter(p.stdout.readline, b"")
        for line in lines_iterator:
            # One exception, if this is the very last report, we will send it
            # through "DONE" command, not "REPORT". So what's happening here
            # is to determine whether this is the last report.
            if cmd_res_dict:
                self.report('REPORT', 'http', cmd_res_dict)
                cmd_res_dict = None
                p_output = line
            else:
                p_output += line
                if line.rstrip() == "}":
                    p_output = self.post_processing(p_output)
                    cmd_res_dict = dict(zip(("status", "stdout", "stderr"), (0, p_output, '')))

        stderr = p.communicate()[1]
        return (p.returncode, p_output, stderr)

    def work(self):
        for item in self.pubsub.listen():
            if item['type'] != 'message':
                continue
            # Convert the string representation of dict to real dict obj
            message = eval(item['data'])
            if message['cmd'] == 'ABORT':
                try:
                    self.last_process.kill()
                except Exception:
                    pass
            else:
                work_thread = threading.Thread(target=agent.process_cmd, args=[message])
                work_thread.daemon = True
                work_thread.start()

    def process_cmd(self, message):
        if message['cmd'] == 'ACK':
            # When 'ACK' is received, means the master node
            # acknowledged the current VM. So stopped sending more
            # "hello" packet to the master node.
            # Unfortunately, there is no thread.stop() in Python 2.x
            self.stop_hello.set()
        elif message['cmd'] == 'EXEC':
            self.last_cmd = ""
            arange = message['data']['active_range']
            my_id = int(self.vm_name[self.vm_name.rindex('I') + 1:])
            if (not arange) or (my_id >= arange[0] and my_id <= arange[1]):
                try:
                    par = message['data'].get('parameter', '')
                    str_par = 'par' if par else ''
                    cmd_res_tuple = eval('self.exec_%s(%s)' % (message['data']['cmd'], str_par))
                    cmd_res_dict = dict(zip(("status", "stdout", "stderr"), cmd_res_tuple))
                except Exception as exc:
                    cmd_res_dict = {
                        "status": 1,
                        "stdout": self.last_cmd,
                        "stderr": str(exc)
                    }
                if self.__class__.__name__ == "KBA_Multicast_Client":
                    self.report('DONE_MC', message['client-type'], cmd_res_dict)
                else:
                    self.report('DONE', message['client-type'], cmd_res_dict)
        else:
            # Unexpected
            print 'ERROR: Unexpected command received!'

class KBA_HTTP_Client(KBA_Client):

    def exec_setup_static_route(self):
        self.last_cmd = KB_Instance.get_static_route(self.user_data['target_subnet_ip'])
        result = self.exec_command(self.last_cmd)
        if (self.user_data['target_subnet_ip'] not in result[1]):
            self.last_cmd = KB_Instance.add_static_route(
                self.user_data['target_subnet_ip'],
                self.user_data['target_shared_interface_ip'])
            return self.exec_command(self.last_cmd)
        else:
            return (0, '', '')

    def exec_check_http_service(self):
        self.last_cmd = KB_Instance.check_http_service(self.user_data['target_url'])
        return self.exec_command(self.last_cmd)

    def exec_run_http_test(self, http_tool_configs):
        self.last_cmd = KB_Instance.run_wrk2(
            dest_path='/usr/local/bin/wrk2',
            target_url=self.user_data['target_url'],
            **http_tool_configs)
        return self.exec_command_report(self.last_cmd)


class KBA_Multicast_Client(KBA_Client):
    def exec_setup_static_route(self):
        self.last_cmd = KB_Instance.get_static_route(self.user_data['target_subnet_ip'])
        result = self.exec_command(self.last_cmd)

        if (self.user_data['target_subnet_ip'] not in result[1]):
            self.last_cmd = KB_Instance.add_static_route(
                self.user_data['target_subnet_ip'],
                self.user_data['target_shared_interface_ip'])
            return self.exec_command(self.last_cmd)
        else:
            return (0, '', '')

    def exec_check_nuttcp_service(self):
        return 0

    def post_processing(self, p_out):
        # Converts form: 'msmaxjitter=59.6045 ms...' into: {'jitter' : 59.6045, 'la'...}
        kmap = {'pkt': 'packets_recv', 'data_loss': 'data_loss', 'drop': 'packets_dropped',
                'megabytes': 'megabytes', 'rate_Mbps': 'mbps', 'msmaxjitter': 'jitter',
                'msavgOWD': 'latency'}  # Format/Include Keys
        try:
            return {kmap[k]: abs(float(v))
                    for (k, v) in [c.split("=")
                    for c in p_out.split(" ")]
                    if k in kmap}
        except Exception:
            return {'error': '0'}



    def exec_multicast_commands_report_helper(self, cmds, timeout=20, trans_recv=0):
        """Start len(cmds) threads to test"""
        queue = multiprocessing.Queue()
        cmd_index = 0
        j = 0
        output = {}

        #  Function for Process  #
        def spawn(cmd, queue):
            p = Popen(cmds[cmd][1].split(" "), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out = p.communicate()[0]
            out = out.replace("\n\n", "\n").split("\n")
            out = out[len(out) / 2]
            queue.put([cmds[cmd][0], out])
        #  End Function  #

        for cmd in cmds:
            multiprocessing.Process(target=spawn, args=(cmd_index, queue)).start()
            cmd_index += 1
        p_err = ""
        try:
            while(j < len(cmds)):
                out = queue.get(True, timeout)
                key = out[0]
                j += 1
                p_out = self.post_processing(out[1].rstrip())
                if key in output:
                    for k in output[key]:
                        output[key][k] += abs(p_out[k])
                else:
                    output[key] = p_out

        except Exception:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            p_err = repr(traceback.format_exception(exc_type, exc_value,
                                                    exc_traceback))

        return (output, str(p_err))

    def exec_mulitcast_commands_report(self, cmds, timeout=20, trans_recv=0):
        """For each batch, pass it off to the helper to start threads.
        exec_multicast_commands_report_helper is blocking.
        """
        j_output = {}
        n_err = 0
        r_err = ""
        for cmd_list in cmds:
            try:
                refresh_clock(self.user_data['ntp_clocks'])
                round_output, err = self.exec_multicast_commands_report_helper(cmd_list,
                                                                               timeout, trans_recv)
                output_key = round_output.keys()[0]
                for key in round_output[output_key]:
                    round_output[output_key][key] /= float(len(cmd_list))
                j_output.update(round_output)
            except Exception:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                err = repr(traceback.format_exception(exc_type, exc_value, exc_traceback))
            finally:
                if err != "":
                    n_err += 1
                    r_err = err
        return (0, json.dumps(j_output), r_err)

    def exec_run_multicast_test(self, nutt_tool_configs):
        """Tests varying nAddresses/ports/bandwidths for a constant packet size.
           Creates Batches of tests to run. One batch of commands is of size nAddresses * nPorts.
           Sends off len(bitrates) list of Batches to exec_mulitcast_commands_report
        """

        commands_list = []
        startAddr = nutt_tool_configs['multicast_address_start'].split(".")
        offset = int(startAddr[-1])
        startAddr = startAddr[:-1]
        startPort = 12000
        max_nAddr = nutt_tool_configs['address_pattern'][-1]
        max_nPort = nutt_tool_configs['port_pattern'][-1]
        duration = nutt_tool_configs['duration']
        server_address = nutt_tool_configs['server_address']
        target_url = self.user_data['target_url'] if server_address == '0.0.0.0' else server_address
        transmit_receive = 't' if server_address == '0.0.0.0' else 'r'
        target_port = 5000 if server_address == '0.0.0.0' else nutt_tool_configs['server_port']

        if transmit_receive == 'r':  # Manually join multicast group...
            for addr_i in range(0, max_nAddr):
                for port_i in range(0, max_nPort):
                    m_port = startPort + ((addr_i) * max_nPort) + (port_i)
                    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                    m_addr = ".".join(startAddr + [str(offset + addr_i)])
                    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
                    mreq = struct.pack("4sl", socket.inet_aton(m_addr), socket.INADDR_ANY)
                    s.bind((m_addr, m_port))
                    s.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
            # Run this garbage test to ensure the listener is in the multicast group.
            comm = KB_Instance.run_nuttcp(dest_path='/usr/local/bin/nuttcp',
                                          target_url=target_url,
                                          target_port=target_port,
                                          multicast_port=12000,
                                          multicast_addr='231.0.0.1',
                                          bitrate=100,
                                          pktsize=1316,
                                          report_interval=60,
                                          transmit_receive='r')

            exec_command(comm.split(" "))

        nReceivers = nutt_tool_configs['receivers']
        pkt_s = nutt_tool_configs['pktsize']
        for index in range(0, len(nutt_tool_configs['address_pattern'])):
            nAddresses = nutt_tool_configs['address_pattern'][index]
            nPorts = nutt_tool_configs['port_pattern'][index]
            for bitrate in nutt_tool_configs['bitrates']:
                commands = []
                for addr_i in range(0, nAddresses):
                    for port_i in range(0, nPorts):
                        m_addr = ".".join(startAddr + [str(addr_i + offset)])
                        m_port = startPort + (addr_i * max_nPort) + (port_i)

                        comm = KB_Instance.run_nuttcp(dest_path='/usr/local/bin/nuttcp',
                                                      target_url=target_url,
                                                      target_port=target_port,
                                                      multicast_port=m_port,
                                                      multicast_addr=m_addr,
                                                      bitrate=bitrate,
                                                      pktsize=pkt_s,
                                                      report_interval=duration,
                                                      transmit_receive=transmit_receive)
                        key = "%d,%d,%d,%f,%d" % (nReceivers, nAddresses, nPorts, bitrate, pkt_s)
                        commands.append([key, comm])
                commands_list.append(commands)

        self.last_cmd = commands_list[-1][-1]
        t_r = 0 if transmit_receive == 't' else duration
        return self.exec_mulitcast_commands_report(commands_list,
                                                   trans_recv=t_r)

class KBA_Storage_Client(KBA_Client):

    def encode_bins(self, p_output):
        p_output = json.loads(p_output)
        p_output['jobs'][0].pop('trim')
        test_list = ['read', 'write']

        for test in test_list:
            histogram = HdrHistogram(1, 5 * 3600 * 1000, 3)
            clat = p_output['jobs'][0][test]['clat']['bins']
            total_buckets = clat['FIO_IO_U_PLAT_NR']
            grp_msb_bits = clat['FIO_IO_U_PLAT_BITS']
            buckets_per_grp = clat['FIO_IO_U_PLAT_VAL']

            for bucket in xrange(total_buckets):
                if clat[str(bucket)]:
                    grp = bucket / buckets_per_grp
                    subbucket = bucket % buckets_per_grp
                    if grp == 0:
                        val = subbucket - 1
                    else:
                        base = 2 ** (grp_msb_bits + grp - 1)
                        val = int(base + (base / buckets_per_grp) * (subbucket - 0.5))
                    histogram.record_value(val, clat[str(bucket)])

            p_output['jobs'][0][test]['clat']['hist'] = histogram.encode()
            p_output['jobs'][0][test]['clat'].pop('bins')
            p_output['jobs'][0][test]['clat'].pop('percentile')

        return json.dumps(p_output)

    def exec_init_volume(self, vol_init_configs):
        self.last_cmd = KB_Instance.init_volume(
            dest_path='/usr/local/bin/fio',
            **vol_init_configs)
        return self.exec_command(self.last_cmd)

    def exec_run_storage_test(self, fio_configs):
        self.last_cmd = KB_Instance.run_fio(
            dest_path='/usr/local/bin/fio',
            name='kb_storage_test',
            **fio_configs)
        return self.exec_command_report(self.last_cmd)

    def post_processing(self, p_out):
        return self.encode_bins(p_out)


class KBA_Server(object):

    def __init__(self, user_data):
        self.user_data = user_data

    def config_nginx_server(self):
        # Generate the HTML file with specified size
        html_size = self.user_data['http_server_configs']['html_size']
        cmd_str = 'dd if=/dev/zero of=/data/www/index.html bs=%s count=1' % html_size
        cmd = cmd_str.split()
        return False if exec_command(cmd) else True

    def start_nginx_server(self):
        cmd = ['sudo', 'service', 'nginx', 'start']
        return exec_command(cmd)

    def start_nuttcp_server(self):
        cmd = ['/usr/local/bin/nuttcp', '-S' '-P5000']
        return exec_command(cmd)

    def start_multicast_listener(self, mc_addrs, multicast_ports, start_address="231.0.0.128"):
        '''Starts Listeners at second /25 (.128).
           These listeners are created when nReceivers > 1.
        '''
        startPort = 12000
        startAddr = start_address.split(".")[:-1]
        start_offset = int(start_address.split(".")[-1])

        #  Thread Function  #
        def spawn_mcl(addr_i, port):
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            m_addr = ".".join(startAddr + [str(addr_i)])
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            ttl = struct.pack('B', 150)
            s.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, ttl)

            mreq = struct.pack("4sl", socket.inet_aton(m_addr), socket.INADDR_ANY)
            s.bind((m_addr, port))
            s.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
            while True:
                d, e = s.recvfrom(10240)

        #  End Function  #

        for addr_i in range(0, mc_addrs):
            for port_i in range(0, multicast_ports):
                m_port = startPort + ((addr_i) * multicast_ports) + (port_i)
                multiprocessing.Process(target=spawn_mcl,
                                        args=(start_offset + addr_i, m_port,)).start()

        while True:
            continue

class KBA_Proxy(object):
    def start_redis_server(self):
        cmd = ['sudo', 'service', 'redis-server', 'start']
        return exec_command(cmd)


if __name__ == "__main__":
    try:
        with open('user-data', 'r') as f:
            user_data = dict(eval(f.read()))
    except Exception as e:
        # KloudBuster starts without user-data
        cwd = 'kloudbuster/kb_server'
        cmd = ['python', 'setup.py', 'develop']
        rc = exec_command(cmd, cwd=cwd)
        if not rc:
            cmd = ['/usr/local/bin/pecan', 'serve', 'config.py']
            sys.exit(exec_command(cmd, cwd=cwd))

    if user_data.get('role') == 'KB-PROXY':
        agent = KBA_Proxy()
        sys.exit(agent.start_redis_server())
    if user_data.get('role').endswith('Server'):
        agent = KBA_Server(user_data)
        if user_data['role'].startswith('Multicast'):
            KB_Instance.add_multicast_route()
            if user_data['n_id'] == 0:
                refresh_clock(user_data.get('ntp_clocks'), force_sync=True)
                agent.start_nuttcp_server()
                while True:
                    refresh_clock(user_data.get('ntp_clocks'))
                    time.sleep(10)
                sys.exit(0)
            else:
                agent.start_multicast_listener(user_data.get('multicast_addresses'),
                                               user_data.get('multicast_ports'),
                                               user_data.get('multicast_listener_address_start'))
        if agent.config_nginx_server():
            sys.exit(agent.start_nginx_server())
        else:
            sys.exit(1)
    elif user_data.get('role').endswith('Client'):
        if user_data['role'].startswith('HTTP'):
            agent = KBA_HTTP_Client(user_data)
        elif user_data['role'].startswith('Multicast'):
            KB_Instance.add_multicast_route()
            refresh_clock(user_data.get('ntp_clocks'), force_sync=True)
            agent = KBA_Multicast_Client(user_data)
        else:
            agent = KBA_Storage_Client(user_data)
        agent.setup_channels()
        agent.hello_thread = threading.Thread(target=agent.send_hello)
        agent.hello_thread.daemon = True
        agent.hello_thread.start()
        agent.work()
    else:
        sys.exit(1)



# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import os

from oslo_config import cfg
from oslo_log import handlers
from oslo_log import log as oslogging

DEBUG_OPTS = [
    cfg.BoolOpt("kb-debug",
                default=False,
                help="Print debug output for KloudBuster only")
]

CONF = cfg.CONF
CONF.register_cli_opts(DEBUG_OPTS)
oslogging.register_options(CONF)

logging.KBDEBUG = logging.DEBUG + 5
logging.addLevelName(logging.KBDEBUG, "KBDEBUG")

CRITICAL = logging.CRITICAL
DEBUG = logging.DEBUG
ERROR = logging.ERROR
FATAL = logging.FATAL
INFO = logging.INFO
NOTSET = logging.NOTSET
KBDEBUG = logging.KBDEBUG
WARN = logging.WARN
WARNING = logging.WARNING

def setup(product_name, logfile=None):
    dbg_color = handlers.ColorHandler.LEVEL_COLORS[logging.DEBUG]
    handlers.ColorHandler.LEVEL_COLORS[logging.KBDEBUG] = dbg_color
    CONF.logging_default_format_string = '%(asctime)s %(levelname)s %(message)s'

    oslogging.setup(CONF, product_name)
    # Adding FileHandler to KloudBuster if logfile is supplied
    if logfile:
        if os.path.exists(logfile):
            os.remove(logfile)
        fmt = logging.Formatter(fmt=CONF.logging_default_format_string)
        hdlr = logging.FileHandler(logfile)
        hdlr.setFormatter(fmt)
        oslogging._loggers[product_name].logger.addHandler(hdlr)

    if CONF.kb_debug:
        oslogging.getLogger(
            project=product_name).logger.setLevel(logging.KBDEBUG)

def getLogger(name="unknown", version="unknown"):
    if name not in oslogging._loggers:
        oslogging._loggers[name] = KloudBusterContextAdapter(
            logging.getLogger(name), {"project": "kloudbuster",
                                      "version": version})

    return oslogging._loggers[name]

def delete_logfile(product_name, filename):
    if filename and os.path.exists(filename):
        os.remove(filename)

class KloudBusterContextAdapter(oslogging.KeywordArgumentAdapter):

    def kbdebug(self, msg, *args, **kwargs):
        self.log(logging.KBDEBUG, msg, *args, **kwargs)



# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

# Module for credentials in Openstack
import getpass
from keystoneauth1.identity import v2
from keystoneauth1.identity import v3
from keystoneauth1 import session
import os
import re

import log as logging

LOG = logging.getLogger(__name__)

class Credentials(object):

    def get_session(self):
        dct = {
            'username': self.rc_username,
            'password': self.rc_password,
            'auth_url': self.rc_auth_url
        }
        auth = None

        if self.rc_identity_api_version == 3:
            dct.update({
                'project_name': self.rc_project_name,
                'project_domain_name': self.rc_project_domain_name,
                'user_domain_name': self.rc_user_domain_name
            })
            auth = v3.Password(**dct)
        else:
            dct.update({
                'tenant_name': self.rc_tenant_name
            })
            auth = v2.Password(**dct)
        return session.Session(auth=auth, verify=self.rc_cacert)

    def get_user_session(self, username, password, tenant_name):
        dct = {
            'username': username,
            'password': password,
            'auth_url': self.rc_auth_url
        }
        auth = None

        if self.rc_identity_api_version == 3:
            dct.update({
                'project_name': tenant_name,
                'project_domain_name': self.rc_project_domain_name,
                'user_domain_name': self.rc_user_domain_name
            })
            auth = v3.Password(**dct)
        else:
            dct.update({
                'tenant_name': tenant_name
            })
            auth = v2.Password(**dct)
        return session.Session(auth=auth, verify=self.rc_cacert)

    def __parse_openrc(self, file):
        export_re = re.compile('export OS_([A-Z_]*)="?(.*)')
        for line in file:
            line = line.strip()
            mstr = export_re.match(line)
            if mstr:
                # get rif of posible trailing double quote
                # the first one was removed by the re
                name = mstr.group(1)
                value = mstr.group(2)
                if value.endswith('"'):
                    value = value[:-1]
                # get rid of password assignment
                # echo "Please enter your OpenStack Password: "
                # read -sr OS_PASSWORD_INPUT
                # export OS_PASSWORD=$OS_PASSWORD_INPUT
                if value.startswith('$'):
                    continue
                # Check if api version is provided
                # Default is keystone v2
                if name == 'IDENTITY_API_VERSION':
                    self.rc_identity_api_version = int(value)

                # now match against wanted variable names
                elif name == 'USERNAME':
                    self.rc_username = value
                elif name == 'AUTH_URL':
                    self.rc_auth_url = value
                elif name == 'TENANT_NAME':
                    self.rc_tenant_name = value
                elif name == "CACERT":
                    self.rc_cacert = value
                elif name == "REGION_NAME":
                    self.rc_region_name = value
                elif name == "PASSWORD":
                    self.rc_password = value
                elif name == "USER_DOMAIN_NAME":
                    self.rc_user_domain_name = value
                elif name == "PROJECT_NAME":
                    self.rc_project_name = value
                elif name == "PROJECT_DOMAIN_NAME":
                    self.rc_project_domain_name = value

    #
    # Read a openrc file and take care of the password
    # The 2 args are passed from the command line and can be None
    #
    def __init__(self, openrc_file, pwd=None, no_env=False):
        self.rc_password = None
        self.rc_username = None
        self.rc_tenant_name = None
        self.rc_auth_url = None
        self.rc_cacert = None
        self.rc_region_name = None
        self.rc_user_domain_name = None
        self.rc_project_domain_name = None
        self.rc_project_name = None
        self.rc_identity_api_version = 2
        success = True

        if openrc_file:
            if isinstance(openrc_file, str):
                if os.path.exists(openrc_file):
                    self.__parse_openrc(open(openrc_file))
                else:
                    LOG.error('Error: rc file does not exist %s', openrc_file)
                    success = False
            else:
                self.__parse_openrc(openrc_file)
        elif not no_env:
            # no openrc file passed - we assume the variables have been
            # sourced by the calling shell
            # just check that they are present
            if 'OS_IDENTITY_API_VERSION' in os.environ:
                self.rc_identity_api_version = int(os.environ['OS_IDENTITY_API_VERSION'])

            if self.rc_identity_api_version == 2:
                for varname in ['OS_USERNAME', 'OS_AUTH_URL', 'OS_TENANT_NAME']:
                    if varname not in os.environ:
                        LOG.warning('%s is missing', varname)
                        success = False
                if success:
                    self.rc_username = os.environ['OS_USERNAME']
                    self.rc_auth_url = os.environ['OS_AUTH_URL']
                    self.rc_tenant_name = os.environ['OS_TENANT_NAME']
                if 'OS_REGION_NAME' in os.environ:
                    self.rc_region_name = os.environ['OS_REGION_NAME']
            elif self.rc_identity_api_version == 3:
                for varname in ['OS_USERNAME', 'OS_AUTH_URL', 'OS_PROJECT_NAME',
                                'OS_PROJECT_DOMAIN_NAME', 'OS_USER_DOMAIN_NAME']:
                    if varname not in os.environ:
                        LOG.warning('%s is missing', varname)
                        success = False
                if success:
                    self.rc_username = os.environ['OS_USERNAME']
                    self.rc_auth_url = os.environ['OS_AUTH_URL']
                    self.rc_project_name = os.environ['OS_PROJECT_NAME']
                    self.rc_project_domain_id = os.environ['OS_PROJECT_DOMAIN_NAME']
                    self.rc_user_domain_id = os.environ['OS_USER_DOMAIN_NAME']
            if 'OS_CACERT' in os.environ:
                self.rc_cacert = os.environ['OS_CACERT']


        # always override with CLI argument if provided
        if pwd:
            self.rc_password = pwd
        # if password not know, check from env variable
        elif self.rc_auth_url and not self.rc_password and success:
            if 'OS_PASSWORD' in os.environ and not no_env:
                self.rc_password = os.environ['OS_PASSWORD']
            else:
                # interactively ask for password
                self.rc_password = getpass.getpass(
                    'Please enter your OpenStack Password: ')
        if not self.rc_password:
            self.rc_password = ""
